#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "main.h"
#include "Graphics.h"
#include "System.h"
#include "math/Complex.h"

#define SCREEN_WIDTH    1280
#define SCREEN_HEIGHT   720
#define MAX_ITERATIONS  100
#define TOLERANCE       1e-6

// the function
Complex f(Complex z)
{
    // z^3 - 1
    static Complex one(1, 0);
    return z * z * z - one;
}

int main(int argc, char* argv[])
{
    // init the window
    gfx.init("Secant fractal", SCREEN_WIDTH, SCREEN_HEIGHT);
    gfx.init2D();
    gfx.clearScreen(Color(0, 0, 0, SDL_ALPHA_OPAQUE));

    static double zoom = 2.0 / (double)SCREEN_HEIGHT;
    static double centerX = (double)SCREEN_WIDTH  / 2.0;
    static double centerY = (double)SCREEN_HEIGHT / 2.0;

    // roots of the function
    Complex roots[3];
    roots[0] = Complex(1, 0);
    roots[1] = Complex(-0.5,  sqrt(3) / 2);
    roots[2] = Complex(-0.5, -sqrt(3) / 2);

    // palette
    static Color    palette[4];
    palette[0] = Color(255, 0, 0);
    palette[1] = Color(0, 255, 0);
    palette[2] = Color(0, 0, 255);
    palette[3] = Color(0, 0, 0);

    // draw the fractal
    for (int y = 0; y < SCREEN_HEIGHT; y++)
    {
        for (int x = 0; x < SCREEN_WIDTH; x++)
        {
            // convert the pixel pos to a complex number
            Complex z[3];

            for (int i = 0; i < 3; i++)
            {
                z[i].re = (x - centerX) * zoom - 0.75 + (2-i) * TOLERANCE;
                z[i].im = (y - centerY) * zoom;
            }

            // by default the pixel is black if the calculations fail
            int palIdx = 3;

            // iterate
            int iter;

            for (iter = 0; iter < MAX_ITERATIONS; iter++)
            {
                z[2] = z[1] - f(z[1]) * (z[1] - z[0]) / (f(z[1]) - f(z[0]));

                z[0] = z[1];
                z[1] = z[2];

                // are we near a root ?
                for (int i = 0; i < 3; i++)
                {
                    Complex delta = z[2] - roots[i];
                    double distance = delta.squareMag();

                    if (distance < TOLERANCE)
                    {
                        palIdx = i;
                        break;
                    }
                }

                if (palIdx != 3)
                    break;
            }

            // draw the pixel
            Color color = palette[palIdx];

            if (palIdx != 3)
            {
                float coef = 1 - log(iter + 1) / 4.0;

                if (coef < 0)
                    coef = 0;

                color.r *= coef;
                color.g *= coef;
                color.b *= coef;
            }
            gfx.setPixel(x, y, color);
        }

        gfx.render();
        sys.processEvents();

        if (sys.isQuitRequested() == true)
            break;
    }

    // wait until we quit
    while (sys.isQuitRequested() == false)
    {
        sys.processEvents();
        sys.wait(16);
    }

    gfx.quit();

    return EXIT_SUCCESS;
}
